/******************************************************************************/
/*   amcc.c -- AMCC S5920 PCI Target Interface driver for custom pcb          */
/*             OCIW CCD-array readout electronics                             */
/*   (c) 2004 Greg Burley (burley@ociw.edu)                                   */
/*            Drew Moore  (acm3035@rit.edu)                                   */
/*                                                                            */
/*   Some of this driver is/was shamelessly based on the code for the         */
/*   AMCC S5933 PCI Matchmaker driver                                         */
/*   (c) 1997 Andrea Cisternino  (acister@pcape1.pi.infn.it)                  */
/*                                                                            */
/*   This program is free software; you can redistribute it and/or            */
/*   modify it under the terms of the GNU General Public License as           */
/*   published by the Free Software Foundation (www.gnu.org)                  */
/*                                                                            */
/******************************************************************************/

/*------------------------------------------------------------------------------
 *  Driver main file -- this file contains the driver code.  The hardware
 *  interface to the PCI bus is an AMCC S5920 chip.  Basically, it bridges the
 *  PCI bus to a local Add-on bus, with support for four pass-thru data regions.
 *  See S5920 PCI Target Interface Data Book from AMCC for details.
 *----------------------------------------------------------------------------*/
#ifndef __KERNEL__
#define __KERNEL__
#endif

#if !defined (CONFIG_PCI)
#error Linux kernel needs PCI support for amcc driver
#endif

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#error Linux kernel version > 2.6.0 required for this driver
#endif

#include <linux/init.h>             /* For __init/__exit/... */
#include <linux/module.h>           /* For module specific items */
#include <linux/kernel.h>           /* For printk/panic/...*/
#include <linux/fs.h>               /* For open, read, write, ioctl */
#include <linux/pci.h>              /* For pci functions */
#include <linux/slab.h>             /* For kmalloc(), kfree() */
#include <linux/ioport.h>           /* For request_region()... */
#include <linux/delay.h>            /* For udelay(), ndelay() */
#include <linux/errno.h>            /* For error reporting */
//#include <linux/sched.h>          /* current and everything */
#include <asm/uaccess.h>            /* For copy_to/from_user() ... */
#include <asm/io.h>                 /* For inl(), outl(), ioremap() ... */
#include "amcc.h"                   /* S5920 definitions */

/*------------------------------------------------------------------------------
 *  Definitions
 *----------------------------------------------------------------------------*/
#define AMCC_NAME               "amcc"
#define AMCC_MAGIC              0x35393230  /* "5920" */
#define AMCC_RD_TIMEOUT         10          /* timeout in microseconds */
#define AMCC_WR_TIMEOUT         500         /* timeout in nanoseconds */
#define AMCC_RD_RETRYS          1000
#define AMCC_WR_RETRYS          50

/* lock bits for concurrent access */
#define AMCC_P2A_LOCK_BIT       0
#define AMCC_A2P_LOCK_BIT       1
#define AMCC_TEST_IRQ_BIT       2

/* return codes for interrupts */
#define AMCC_IRQ_READ_OK        1
#define AMCC_IRQ_WRITE_OK       2
#define AMCC_IRQ_MASTER_ABORT   3
#define AMCC_IRQ_TARGET_ABORT   4
#define AMCC_IRQ_OMB_EMPTY      5
#define AMCC_IRQ_IMB_FULL       6

#ifndef PCI_VENDOR_ID_AMCC
#define PCI_VENDOR_ID_AMCC      0x10e8
#endif

#define PCI_DEVICE_ID_S5920     0x5920

#define AMCC_MAJOR              125     /* Set to 0 for dynamic major number */
#define AMCC_FIFO_SIZE          65536   /* FIFO is 65536 16-bit words */
#define P2A_BUFFER_SIZE         4096    /* Size in bytes */
#define A2P_BUFFER_SIZE         131072  /* Size in bytes */

/*------------------------------------------------------------------------------
 *  Structures
 *----------------------------------------------------------------------------*/

/* Buffer info -- holds data for Pass-Thru buffer handling. */
struct amcc_buffer {
    volatile u_int copy_size;
    volatile u_int bytes_done;
    volatile u_int size;
    char *buf;
};

/* Driver private info -- holds global data for a single PCI board. */
struct amcc_priv {
/*  struct pci_dev *pcidev; */
    u32 opreg_addr;             /* operation registers address BAR[0] */
    u32 opreg_len;              /* operation registers size */
    int opreg_type;             /* MEM or I/O operation registers */
    int ptnum;                  /* number of pass-thru regions */
    u_long pt1_phys_addr;       /* Pass-thru region 1 - BAR[1] */
    u_long pt1_size;            /* Pass-thru region 1 size */
    void*  pt1_virt_addr;       /* Virtual address after remapping */
    int irq;                    /* interrupt number */
};

/*------------------------------------------------------------------------------
 *  Static variables -- these can be seen by every process using the device.
 *  Use only one board per PC, please.
 *----------------------------------------------------------------------------*/
struct amcc_priv dev;           /* Private device data */
struct amcc_buffer p2a;         /* P2A buffer control struct */
struct amcc_buffer a2p;         /* A2P buffer control struct */
static u_long lock = 0UL;       /* generic lock dword */

static int amcc_major = AMCC_MAJOR;
static int amcc_minor = 0;
static struct class *amcc_class;

/*------------------------------------------------------------------------------
 *  Function prototypes and inline functions
 *----------------------------------------------------------------------------*/
static int amcc_probe (struct pci_dev *, const struct pci_device_id *);
static void amcc_remove (struct pci_dev *);
static int amcc_open (struct inode *, struct file *);
static int amcc_close (struct inode *, struct file *);
static ssize_t amcc_read (struct file *, char *, size_t, loff_t *);
static ssize_t amcc_write (struct file *, const char *, size_t, loff_t *);
static int amcc_ioctl (struct inode *, struct file *, unsigned int, unsigned long);
//static irqreturn_t amcc_irq_handler (int, void *, struct pt_regs *);

static struct file_operations amcc_fops =
{
    .owner = THIS_MODULE,
    .open = amcc_open,
    .release = amcc_close,
    .read = amcc_read,
    .write = amcc_write,
    .ioctl = amcc_ioctl,
};

/* Functions for io-mapped access to operations registers in BAR[0].        *
 * Memory-mapped access to operations registers commented out for now --    *
 * don't need this on x86.                                                  */
static inline __u32
amcc_read_opreg (int reg)
{
    return ((__u32) inl((unsigned int)(dev.opreg_addr + reg)));
/*  return (readl ((amcc_dev.opreg_addr + reg))); */
}

static inline void
amcc_write_opreg (__u32 value, int reg)
{
    outl(value, (unsigned int)(dev.opreg_addr + reg));
/*  writel (value, (amcc_dev.opreg_addr + reg)); */
}

/*------------------------------------------------------------------------------
 *  Module initialization and remove functions.  
 *----------------------------------------------------------------------------*/
static int
amcc_probe (struct pci_dev* pcidev, const struct pci_device_id* amcc_device_id)
{
    extern struct amcc_priv dev;
    int rc;
    int j;

    /* Register the private data area per board. Attempt to enable the device.
     * We don't use the private area in the way it's intended, yet */
    pci_set_drvdata(pcidev, &dev);
    rc = pci_enable_device(pcidev);
    if (rc != 0) {
        printk (KERN_WARNING "amcc_probe: device enable failed \n");
        return -EIO;
    }

    /* Register as a char() device with major number = 125 */
    rc = register_chrdev(amcc_major, "amcc", &amcc_fops);
    if (rc < 0) {
        printk (KERN_WARNING "amcc_probe: unable to get major number\n");
        return -EIO;
    }
    if (amcc_major == 0) amcc_major = rc; /* dynamic */
    printk (KERN_DEBUG "amcc_probe: major = %d\n", amcc_major);

    amcc_class = class_create(THIS_MODULE, "amcc");
    if (IS_ERR(amcc_class)) {
        printk(KERN_ERR "Error creating amcc class.\n");
        return -EIO;
    }

     /* Add device with minor number = 0  */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,26)    /* Fedora8 kernel 2.6.26.x */
    device_create(amcc_class, NULL,  MKDEV(amcc_major,amcc_minor),
        "amcc%u", amcc_minor);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)    /* Fedora10 kernel 2.6.27.x */
    device_create(amcc_class, NULL,  MKDEV(amcc_major, amcc_minor),
        NULL, "amcc%u", amcc_minor);
#endif

    /* Allocate buffers for Add-on to/from PCI bus transfers */
    p2a.buf = (char *) kmalloc(P2A_BUFFER_SIZE, GFP_KERNEL);
    p2a.size = P2A_BUFFER_SIZE;

    a2p.buf = (char *) kmalloc(A2P_BUFFER_SIZE, GFP_KERNEL);
    a2p.size = A2P_BUFFER_SIZE;

    if ((p2a.buf == NULL) || (a2p.buf == NULL)) {
        printk (KERN_WARNING "amcc_probe: no memory for Pass-thru buffers\n");
        return -ENOMEM;
    }

    /* Reserve the BAR regions of the PCI device  */
    rc = pci_request_regions(pcidev, "amcc");
    if (rc == -EBUSY) {
        printk (KERN_WARNING "amcc_probe: IO request failed\n");
        return -EIO;
    }

    /* Retrieve from PCI_BASE_ADDRESS_0 the start address and size of
     * the S5920 operations registers. */
    dev.opreg_addr = pci_resource_start(pcidev, 0);
    dev.opreg_len = pci_resource_len(pcidev, 0);

    /* Check for I/O mapped or Memory mapped Operation Registers */
    if (pci_resource_flags(pcidev, 0) & IORESOURCE_IO) {
        dev.opreg_type = PCI_BASE_ADDRESS_SPACE_IO;
        printk (KERN_DEBUG "amcc_probe: PCI BASE ADDRESS SPACE IO\n");
    }
    else if (pci_resource_flags(pcidev, 0) & IORESOURCE_MEM) {
        dev.opreg_type = PCI_BASE_ADDRESS_SPACE_MEMORY;
        printk (KERN_DEBUG "amcc_probe: PCI BASE ADDRESS SPACE MEM\n");
    }

    /* For Pass-thru region 1, extract the address and size of BAR         *
     * PCI_BASE_ADDRESS_1, then remap into memory space.  Ignore other     *
     * S5920 pass-thru regions (PT[2] thru PT[4]) which are not used here. */
    dev.pt1_size = pci_resource_len(pcidev, 1);
    dev.pt1_phys_addr = pci_resource_start(pcidev, 1);
    dev.pt1_virt_addr = ioremap(dev.pt1_phys_addr, dev.pt1_size);
    printk (KERN_DEBUG "amcc_probe: PT[%d] Size %8d  Phys %08x  Virt %08x\n",
        1, (int)(dev.pt1_size), (int)(dev.pt1_phys_addr), (int)(dev.pt1_virt_addr));

    /* Board reset -- set some defaults */
    amcc_write_opreg(RCR_RESET_MBFLAGS, AMCC_OP_REG_RCR);
    amcc_write_opreg(RCR_RESET_ADDON, AMCC_OP_REG_RCR);
    for (j=0; j<500; j++) udelay(1000);
    amcc_write_opreg((u32) 0, AMCC_OP_REG_RCR);

    amcc_write_opreg((u32) 0, AMCC_OP_REG_ICSR);
    amcc_write_opreg((u32) 0, AMCC_OP_REG_PTCR);

    /* Output some debug info -- at power-up:                         *
     * ICSR 0x00000C0C RCR 0x00000000 MBEF 0x80000000 PTCR 0x80808080 */
    printk (KERN_DEBUG "amcc_probe: ICSR: 0x%08x  RCR : 0x%08x\n",
        amcc_read_opreg(AMCC_OP_REG_ICSR), amcc_read_opreg(AMCC_OP_REG_RCR));
    printk (KERN_DEBUG "amcc_probe: MBEF: 0x%08x  PTCR: 0x%08x\n",
        amcc_read_opreg(AMCC_OP_REG_MBEF), amcc_read_opreg(AMCC_OP_REG_PTCR));
    return 0;
}

static void
amcc_remove (struct pci_dev* pcidev)
{
    extern struct amcc_priv dev;
    struct amcc_priv *priv;

    /* Free all the resources allocated by the driver */
    priv = pci_get_drvdata(pcidev);
    unregister_chrdev(amcc_major, "amcc");

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,26)    /* Fedora8 kernel 2.6.26.x */
    class_device_destroy(amcc_class,MKDEV(amcc_major,amcc_minor));
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)    /* Fedora10 kernel 2.6.27.x */
    device_destroy(amcc_class, MKDEV(amcc_major, amcc_minor));
#endif

    class_destroy(amcc_class);
/*  free_irq (amcc_dev.irq, NULL);  */

    if (p2a.buf) kfree(p2a.buf);
    if (a2p.buf) kfree(a2p.buf);

    iounmap((void *) (dev.pt1_virt_addr));

    pci_release_regions(pcidev);
    pci_disable_device(pcidev);
}


/*------------------------------------------------------------------------------
 *  Standard module open() and close() functions.  
 *----------------------------------------------------------------------------*/
static int
amcc_open (struct inode *inode, struct file *file)
{
    printk (KERN_DEBUG "amcc_open : major number %d  minor number %d\n",
        imajor(inode), iminor(inode));
    return 0;
}


static int
amcc_close (struct inode *inode, struct file *file)
{
    printk (KERN_DEBUG "amcc_close: close\n");
    return 0;
}

/*------------------------------------------------------------------------------
 *  Module read() entry point.
 *  Read from the Add-on bus to PCI bus. Fill a kernel buffer with the 
 *  specified number of bytes.  Hardware specific.
 *----------------------------------------------------------------------------*/
static ssize_t
amcc_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
    extern struct amcc_priv dev;
    extern struct amcc_buffer a2p;
    extern u_long lock;
    ssize_t  total_bytes_read = 0;
    u32 *ptr, *pcnt, imb;
    u16 *rbuf;
    int j, k=0; 
    int nwords, retrys;

    if (count == 0) {
        return 0;
    }

    /* Is the Pass_Thru + FIFO available? */
    if (test_and_set_bit(AMCC_A2P_LOCK_BIT, &lock)) {
        printk (KERN_DEBUG "amcc: Read PT+FIFO already locked\n");
        return -EBUSY;
    }

    /* Set up to read the device into the a2p buffer */
    ptr = (u32 *)(dev.pt1_virt_addr);
    pcnt = (u32 *)(dev.pt1_virt_addr);      /* should be dev.pt2_virt_addr */
    rbuf = (u16 *)a2p.buf;

    a2p.bytes_done = 0;
    a2p.copy_size = min(count, a2p.size);


/* Read until done, use 'burst' mode if possible.                              *
 *                                                                             *
 * The device hardware includes a FIFO. The PAE (programmable-almost-empty)    *
 * flag indicates 127 words (or more) in the FIFO.  Hardware-wise, the FIFO    *
 * EF and PAE are connected to the AMCC chip incoming-mailbox-reg (IMB).       *

 * Check the FIFO empty flag. If the FIFO is not ready, wait a short while.    *
 * If the FIFO is ready, check the PAE flag. Do a burst read if possible,      *
 * or single read otherwise.                                                   */ 

    j = 0;
    retrys = 0;

    while (a2p.bytes_done < a2p.copy_size  && retrys < AMCC_RD_RETRYS){
        imb = amcc_read_opreg(AMCC_OP_REG_IMB);
        if (imb & IMB_RX_RDY){
            if (imb & IMB_RX_PAE){
                nwords = min((int)((a2p.copy_size-a2p.bytes_done)/2), 127);
/*              nwords = (u32) readl ((u32 *)(pcnt) );  EPLD counter -- not implemented  */
                for (k=0; k<nwords; k++){
                    rbuf[j] = (u16) ioread16(&ptr[k]);
                    a2p.bytes_done += 2;
                    j++;
                }
            }
            else {
                rbuf[j] = (u16) ioread16(&ptr[k]);
                a2p.bytes_done += 2;
                j++;
            }
            retrys = 0;
        }
        else {
            if (a2p.copy_size == 2)break;
            retrys += 1;
/*          if (file->f_flags & O_NONBLOCK) break; */       
            udelay(AMCC_RD_TIMEOUT);            /* block for a short while */
        }
    }

    if(copy_to_user(buf, a2p.buf, a2p.bytes_done))
        return -EFAULT;
    total_bytes_read += a2p.bytes_done;
    count -= a2p.bytes_done;

    clear_bit(AMCC_A2P_LOCK_BIT, &lock);
    return total_bytes_read;
}

/*------------------------------------------------------------------------------
 *  Standard write() entry point.
 *  Write to the PCI-to-Add-on FIFO. Copy the user buffer into our (smaller) 
 *  kernel buffer and then perform a transfer. Hardware specific.
 *----------------------------------------------------------------------------*/
static ssize_t
amcc_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    extern struct amcc_priv dev;
    extern struct amcc_buffer p2a;
    extern u_long lock;
    ssize_t total_bytes_written = 0;
    u32 *ptr;
    u16 *wbuf;
    int j,k;

    if (test_and_set_bit(AMCC_P2A_LOCK_BIT, &lock)) {
        printk (KERN_DEBUG "amcc_write: Pass_Thru + FIFO already locked\n");
        return -EBUSY;
    }

    /* Can we check to see if the Pass-Thru + FIFO is empty? */

    /* Set up to write to the device from the p2a buffer */
    ptr = (u32 *)(dev.pt1_virt_addr);
    wbuf = (u16 *)p2a.buf;

    while (count > 0){
        p2a.bytes_done = 0;
        p2a.copy_size = min (count, p2a.size);
        if (copy_from_user(p2a.buf, buf, p2a.copy_size))
            return -EFAULT;
        for (j=0; j<p2a.copy_size/2; j++){
            for (k=0; k < AMCC_WR_RETRYS; k++) {
                ndelay(AMCC_WR_TIMEOUT);
                if (!(amcc_read_opreg(AMCC_OP_REG_IMB) & IMB_TX_BUSY))
                    break;
            }
            iowrite32((u32)wbuf[j], &ptr[j]);
/*          udelay(5);   */                           /* necessary?? */
            p2a.bytes_done += 2;
        }
        count -= p2a.bytes_done;
        total_bytes_written += p2a.bytes_done;
    }
    clear_bit (AMCC_P2A_LOCK_BIT, &lock);
    return total_bytes_written;
}

/*------------------------------------------------------------------------------
 *  Standard module ioctl() entry point.
 *----------------------------------------------------------------------------*/
static int
amcc_ioctl (struct inode *inode, struct file *file,
    unsigned int cmd, unsigned long arg)
{
    extern struct amcc_priv dev;
    extern u_long lock;
    u32 *addr = (u32 *) arg;
    u32 count=0;
    int rc = 0L;
    int j;

    switch (cmd) {
    case AMCC_RESET:            /* reset the whole board */
        amcc_write_opreg (RCR_RESET_MBFLAGS, AMCC_OP_REG_RCR);
        amcc_write_opreg (RCR_RESET_ADDON, AMCC_OP_REG_RCR);
        amcc_write_opreg ((u32) 0, AMCC_OP_REG_RCR);
        lock = 0L;
        break;

    case AMCC_RESET_ADDON:      /* reset the Add-on bus */
        amcc_write_opreg (RCR_RESET_ADDON, AMCC_OP_REG_RCR);
        for (j=0; j<100; j++)udelay(1000);
        amcc_write_opreg ((u32) 0, AMCC_OP_REG_RCR);
        lock = 0L;
        break;

    case AMCC_RESET_FIFO:       /* reset the PCI card external FIFO */
        for (j=0; j<AMCC_FIFO_SIZE; j++) {
            if (!(amcc_read_opreg (AMCC_OP_REG_IMB) & IMB_RX_RDY))
                break;
            (void)ioread32(dev.pt1_virt_addr);
            count +=2;
        }
        rc = put_user(count, addr);
        break;

    case AMCC_GET_ICSR:         /* read ICSR register */
        rc = put_user(amcc_read_opreg (AMCC_OP_REG_ICSR), addr);
        break;

    case AMCC_GET_RCR:          /* read RCR register */
        rc = put_user(amcc_read_opreg (AMCC_OP_REG_RCR), addr);
        break;

    case AMCC_GET_PTCR:         /* read PTCR register */
        rc = put_user(amcc_read_opreg (AMCC_OP_REG_PTCR), addr);
        break;

    case AMCC_GET_MBEF:         /* read mailbox MBEF register */
        rc = put_user(amcc_read_opreg (AMCC_OP_REG_MBEF), addr);
        break;

    default:
        rc = -EINVAL;
        break;
    }
    return rc;
}

/*------------------------------------------------------------------------------
 *  Standard support entries.
 *----------------------------------------------------------------------------*/
MODULE_LICENSE("GPL");
MODULE_AUTHOR("(c) 2004 Greg Burley  www.ociw.edu");
MODULE_DESCRIPTION("Driver for CCD pci interface pcb");

static struct pci_device_id id_table[] __devinitdata = {
    {PCI_VENDOR_ID_AMCC, PCI_DEVICE_ID_S5920, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {0,}    // end-of-list
};

MODULE_DEVICE_TABLE(pci, id_table);

static struct pci_driver amcc_driver = {
    .name = "amcc",
    .id_table = id_table,
    .probe = amcc_probe,
    .remove = __devexit_p(amcc_remove)
};

int __init
amcc_init (void)
{
    int rc = 0;

    printk (KERN_INFO "amcc_init: AMCC S5920 Rev 2.0a for 2.6 kernel \n");
    rc = pci_register_driver(&amcc_driver);
    if (rc < 0) printk (KERN_ERR "amcc_init: Unable to register PCI driver \n");
    return rc;
}

void __exit
amcc_exit (void)
{
    printk (KERN_INFO "amcc_exit: AMCC S5920 Goodbye\n");
    pci_unregister_driver(&amcc_driver);
}

module_init(amcc_init);
module_exit(amcc_exit);
